/*
	Hovedklasse som tar seg opptegning av
	vertex'en til terrenget og frustum culling
*/

import gl4java.GLContext;
import gl4java.awt.GLCanvas;
import gl4java.utils.glut.*;
import gl4java.utils.textures.*;

import java.awt.event.*;
import java.awt.*;
import java.awt.image.*;
import java.util.*;

public class Terreng extends GLCanvas
{
	//Variabler for  lagre innformasjon om hydekart
	int SIZE;
	float stride;
	
	//Array til  lagre filnavn p teksturer
	String[] text;
	
	float down;

	//Variabler til  bevege kamera
	float yrot = 0.0f;
	float	xrot = 0.0f;
	float rotz = 0.0f;
	float xpos = 0.0f;
	float ypos = 0.0f;
	float lookupdown = 0.0f;
	
	float heightdiff;
	float wavediff;
	float wpos;
	
	boolean move = false;
	boolean renderwater = false;
	
	GLUTFunc glut = null;
	
	HeightField hf;
	HeightField waterhf;
	
	HeightImageLoader heightImage;
	HeightImageLoader waterHeight;
	
	Smoothing smooth;
	
	Normals normal;
	Normals normalWater;
	
	Texture myTex;
	boolean mytex;
	
	//Array til  lagre vertex'er for normaler for vann
	float[][][] normalsWater;
	
	float[][][] textureSky; 
	
	Sphere sp = new Sphere();
	//Array til  lagre vertex'er for himmel
	float[][][] sphere;

	//Array til  lagre vertex'er for normaler for terreng
	float[][][] normals;
	
	float[][][] texture;
	float[][][] texturewater;
	
	//Array til  lagre egenlagd tekstur
	float[] madeTex;
	
	//Tekstur objekt
	TextureLoader texLoader = new AWTTextureLoader(this, gl, glu);
	
	int[] texName = {0};
	
	int watertext = 0;
	int groundtext = 0;
	int skytext = 0;
	
	//Array til  lagre displaylister for terreng
	int[] terreng;
	//Array til  lagre displaylister for terreng
	int[] waterDisplay;
	int skyDisplay;
	
	boolean take;
	boolean rendersky = true;
	
	//Bakgrunsfarger
	float[] bakgrunn;
	
	//Array til  lagre frustum
	float frustum[][] = new float[6][4];

	//Konstruktr for  sette opp terrenget med uten vann
	public Terreng(int w, int h, Image img, float stride, float heightdiff, float skyheight, float skydown, String text[], float[] repeat, boolean madetex, boolean take, float[] bakgrunn)
	{
		super(w, h);
		
		this.take = take; 
		//Bakgrunsfarger
		this.bakgrunn = bakgrunn;
		//Sett strrelse variabler  
		SIZE = img.getWidth(this);
		this.stride = stride;
		this.heightdiff = heightdiff;
		
		//Last hyde flate
		hf = new HeightField(SIZE, stride, repeat[0]);
		
		//Last hyde forskjeller
		heightImage = new HeightImageLoader(hf);
		heightImage.loadImage(img);
		heightImage.setData(heightdiff, -skydown);
		
		//Glatt terreng
		smooth = new Smoothing(hf);
		for (int i = 0; i <2; i++) smooth.smooth(0.75f);
		
		//Normaler for terreng
		normal = new Normals(hf);
		normal.calcInterNormals();
		normals = normal.getNormals();
		
		//Tekstur koordinater for terreng
		texture = hf.getTeksturKoor();
		
		//F texture for terreng
		myTex = new Texture(hf, heightdiff-skydown, 0.20f, 0.30f);
		madeTex = myTex.getTexture();
		mytex = madetex;
		
		//Sett opp himmel
		sp.setData(SIZE, stride, 1.0f, skyheight);
		sphere = sp.getSphere();
		textureSky = sp.getTextureKoord(repeat[2]);
		
		this.text = text;

		//Kameraets posisjon
		yrot = (SIZE*stride)/2;
		xrot = (SIZE*stride)/2;
	}

	//Konstruktr for  sette opp terrenget med alle komponenter (himmel, vann, bakke)
	public Terreng(int w, int h, Image img, Image water, float wpos, float stride, float wavediff, float heightdiff, float skyheight, float skydown, String[] text, float[] repeat, boolean madetex, boolean take, float[] bakgrunn)
	{
		this(w, h, img, stride, heightdiff, skyheight, skydown, text, repeat, madetex, take, bakgrunn);
		
		//Velg at vann skal bli rendret
		renderwater = true;
		
		//Vannet's hyde over bakken
		this.wpos = wpos;
		//Blge strrelse
		this.wavediff = wavediff;
		//Lag blge omrde
		waterhf = new HeightField(SIZE, stride, repeat[1]);	
		texturewater = waterhf.getTeksturKoor();
		
		//Last blger
		waterHeight = new HeightImageLoader(waterhf);
		waterHeight.loadImage(water);
		
		//Blgenormaler
		normalWater = new Normals(waterhf);
		waterHeight.setData(this.wavediff, wpos+(-skydown));
		normalWater.calcInterNormals();
		normalsWater = normalWater.getNormals();
		
		//Hvor langt ned himmelen skal vre under terrenget
		down = skydown;
	}
	
	//Sette opp openGL context
	public void preInit()
	{
		doubleBuffer = true;
		stereoView = false;
		accumSize = 8;	
	}
	
	//Sett opp faste variabler for fremvisning
	public void init()
	{
		requestFocus();
		addKeyListener(new keyHandler());
		
		glut = new GLUTFuncLightImpl(gl, glu);
		
		reshape(getParent().getSize().width, getParent().getSize().height);
		
		//sett opp lyskilder
		float ambient[] = {0.4f, 0.4f, 0.4f, 1.0f}; //tilstedevrende lys uten lyskilde
		float diffuse[] = {1.0f, 1.0f, 1.0f, 1.0f}; //lys med fast lyskilde
		float position[] = {0.0f, 0.0f, heightdiff*3.0f, 0.0f}; //lyskildes posisjon
		
		//sett lyskilder p plass
		gl.glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
		gl.glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);

		gl.glLightfv(GL_LIGHT0, GL_POSITION, position);
	
		//sl lyskilder p
		gl.glEnable(GL_LIGHT0);
		gl.glEnable(GL_LIGHTING);
		
		//Last vanntekstur
		watertext = text(text[0]);
		//Last himmeltekstur
		skytext = text(text[1]);
		//Hvis egendefinert tekstur last egentekst
		if (mytex) groundtext = groundtext(); 
		//Hvis ikke last terrengetekstur
		else groundtext = text(text[2]);
	
		//Last tke
		if (take) take();

		//Hvis rendre himmel
		if (rendersky)
		{
			//Lag himmel displayliste
			skyDisplay = gl.glGenLists(((SIZE/32)*(SIZE/32))+1);
			gl.glNewList(skyDisplay, GL_COMPILE);
				setMaterialSky();
				drawSky();
			gl.glEndList();
		}
		
		//Generere terrengedisplaylister
		int nr = generateTerrengDisplay();
		//Generer vann displaylister
		generateWaterDisplay(nr);
		
		//gl.glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,2);

		gl.glEnable(GL_NORMALIZE);

		//glatt ut skygger p tegning
		gl.glShadeModel(GL_SMOOTH);
		
		//dypde sortering av flater
		gl.glEnable(GL_DEPTH_TEST);
		gl.glDepthFunc(GL_LESS);
		
		//Set bakgrundsfarge
		gl.glClearColor(bakgrunn[0], bakgrunn[1], bakgrunn[2], 1.0f); //(R,G,B)
	}
	
	
	//Metode som generere displaylister med 32*32 deler av terrenget
	private int generateTerrengDisplay()
	{
		int nr = 1;
		int x = 0;
		int y = 0;
		int end = 0;
		
		terreng = new int[(SIZE/32)*(SIZE/32)];
		/*
			Gr igjennom alle vertex'ene for terreng-hydekartet
			og genererer displaylister med 32*32 biter av terrenget
		*/
		for (int i = 0; i < SIZE/32; i++)
		{
			for(int j = 0; j < SIZE/32; j++)
			{
				terreng[nr-1] = gl.glGenLists(nr);
				gl.glNewList(terreng[nr-1], GL_COMPILE);
					setMaterialGround();
					drawTerrengDel(x,x+31,y, y+31);
				gl.glEndList();
				y += 31;
				nr++;
			}
			x += 31;
			y = 0;
		}
		return nr;
	}
	
	//Metode som generere displaylister med 32*32 deler av vannet 
	private void generateWaterDisplay(int nr)
	{
		int wnr = 0;
		int x = 0;
		int y = 0;
		int end = 0;
		
		waterDisplay = new int[(SIZE/32)*(SIZE/32)];
		/*
			Gr igjennom alle vertex'ene for vann-hydekartet
			og genererer displaylister med 32*32 biter av vannet
		*/
		for (int i = 0; i < SIZE/32; i++)
		{
			for(int j = 0; j < SIZE/32; j++)
			{
				waterDisplay[wnr] = gl.glGenLists(nr);
				gl.glNewList(waterDisplay[wnr], GL_COMPILE);
					setMaterialWater();
					drawWaterDel(x,x+31,y, y+31);
				gl.glEndList();
				y += 31;
				wnr++;
				nr++;
			}
			x += 31;
			y = 0;
		}
	}
	
	//Tegn opp og vis tegninger, blir kalt hver gang forandring inntreffer
	public void display()
	{
		
		
		//Sett current object
		if (glj.gljMakeCurrent() == false)
			return;

		//reshape(getParent().getSize().width, getParent().getSize().height);
	
		float xtrans = xrot;						//Variable brukt til  flytte p x-akse
		float ytrans = yrot;						//Variable brukt til  flytte p y-akse
		float sceneroty = 360.0f - rotz; //Variabel brukt til  rotere scene
	
		//tm buffer for data
		gl.glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		gl.glClear(GL_ACCUM_BUFFER_BIT);

		//Last identitetsmatrise, tm gammel info
		gl.glLoadIdentity();

		gl.glRotatef(lookupdown,1.0f,0.0f,0.0f);					// Roter landskapet opp og ned
		gl.glRotatef(sceneroty,0.0f,1.0f,0.0f);					// Roter landskapet
			
		//plaser ye og bestem hvor fokus skal vre
		glu.gluLookAt(xtrans, ytrans, hf.getHeight((int)(xtrans/stride), (int)(ytrans/stride))-down,  //yets posisjon
						xtrans/stride, ytrans/stride, heightdiff/2, 				//yet ser p 
						0.0f, 0.0f, 1.0f);				//yet er oppe
	
		gl.glPushMatrix();

		extractFrustum();
			
		int nr = 0;
		int x = 0;
		int y = 0;
		//G gjennom alle display listene
		for (int i = 0; i < SIZE/32; i++)
		{
			for(int j = 0; j < SIZE/32; j++)
			{
				//Hvis midten av en terrengdel er innfor frustum
				if (cubeInFrustum(hf.getLength(x+16,y+16), hf.getWidth(x+16, y+16), hf.getHeight(x+16, y+16), 32*stride))
				{
					//Tegn terrengdel
					gl.glCallList(terreng[nr]);
				}
				nr++;
				y += 32;
			}
			x += 32;
			y = 0;
		}
			
		//Hvis vannet skal tegnes opp
		if (renderwater)
		{
			setMaterialWater();		
			int wnr = 0;
			int l = 0;
			int k = 0;
			//G gjennom alle display listene
			for (int i = 0; i < SIZE/32; i++)
			{
				for(int j = 0; j < SIZE/32; j++)
				{
					//Hvis midten av en vanndel er innfor frustum
					if (cubeInFrustum(waterhf.getLength(l+16,k+16), waterhf.getWidth(l+16, k+16), waterhf.getHeight(l+16, k+16), 32*stride))
					{
						//Tegn opp vanndel
						gl.glCallList(waterDisplay[wnr]);
					}
					wnr++;
					k += 32;
				}
				l += 32;
				k = 0;
			}
		}
			
		//Flytt origo til midt p terrenget slik at himmeltoppen kommer p midten
		gl.glTranslatef((float)SIZE/2*stride, (float)SIZE/2*stride, 0.0f);
		
		//Hvis himmel skal tegnes opp
		if (rendersky)
		{
			//Tegn opp himmel
			gl.glCallList(skyDisplay);
		}
		
		gl.glPopMatrix();
	
		//rydd opp
		glj.gljSwap();
		glj.gljFree();
	}
	
	//Intreffer ved forandringer av vindustrrelse
	public void reshape(int w, int h)
	{
		//Strrelse p bilde
		gl.glViewport(0,0, w, h);
		
		gl.glMatrixMode(GL_PROJECTION);
		gl.glLoadIdentity();
		
		//yets perspective p objektet
		glu.gluPerspective(50.0, (float) w/h, 1.0, (SIZE*stride)*3);

		gl.glMatrixMode(GL_MODELVIEW);	
	}

	//Metode som tegner opp tke
	private void take()
	{
			//Rendrer ikke himmel
			rendersky = false;
			//Slr p tke
			gl.glEnable(GL_FOG);
			gl.glHint(GL_FOG_HINT, GL_NICEST);
		
			//Setter tkefarge lik bakgrunsfarge
			float color[] = {bakgrunn[0], bakgrunn[1], bakgrunn[2], 1.0f};
			//Velger hvordan tken skal vises
			gl.glFogi(GL_FOG_MODE, GL_EXP);
			
			gl.glFogfv(GL_FOG_COLOR, color);
			//Setter tkens tykhet
			gl.glFogf(GL_FOG_DENSITY, 0.00001f);
			gl.glFogf(GL_FOG_START, 0.0f);
			gl.glFogf(GL_FOG_END, stride);
	}
	
	
	//Metode som teger opp himmel
	private void drawSky()
	{
		//Slr p tekstur
		gl.glEnable(GL_TEXTURE_2D);
		//Binder tekstur
		gl.glBindTexture(GL_TEXTURE_2D, skytext);
		//Velger hvordan tekstur skal blandes med bakgrunn
		gl.glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL);
		//Tegner opp vertex'ene til himmel
		for (int x = 0; x < 36; x++)
		{
			gl.glBegin(GL_TRIANGLE_STRIP);
			for (int y = 0; y < 37; y++)
			{	
					gl.glTexCoord2f(textureSky[x][y][0], textureSky[x][y][1]); gl.glVertex3f(sphere[x][y][0], sphere[x][y][1], sphere[x][y][2]);
					gl.glTexCoord2f(textureSky[x+1][y][0], textureSky[x+1][y][1]); gl.glVertex3f(sphere[x+1][y][0], sphere[x+1][y][1], sphere[x+1][y][2]);
			}
			gl.glEnd();
		}
		//Slr av tekstur
		gl.glDisable(GL_TEXTURE_2D);
	}

	//Metode som tegner opp definert del av terreng
	private void drawTerrengDel(int x1, int x2, int y1, int y2)
	{
		//Slr p tekstur
		gl.glEnable(GL_TEXTURE_2D);
		//Binder tekstur
		gl.glBindTexture(GL_TEXTURE_2D, groundtext);
		//Velger hvordan tekstur skal blandes med bakgrunn
		gl.glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
		//Tegner vertex'ene til terrenget
		for (int x = x1; x <= x2; x++)
		{
			gl.glBegin(GL_TRIANGLE_STRIP);
			for (int y = y1; y <= y2; y++)
			{
					gl.glNormal3f(normals[x][y][0], normals[x][y][1], normals[x][y][2]);
					gl.glTexCoord2f(texture[x][y][0], texture[x][y][1]); gl.glVertex3f(hf.getLength(x,y), hf.getWidth(x, y), hf.getHeight(x, y));
					gl.glTexCoord2f(texture[x+1][y][0], texture[x+1][y][1]); gl.glVertex3f(hf.getLength(x+1,y), hf.getWidth(x+1, y), hf.getHeight(x+1, y));	
					
			}
			gl.glEnd();
		}
		//Slr av tekstur
		gl.glDisable(GL_TEXTURE_2D);
	}

	//Metode som tegner opp definert del av vann
	private void drawWaterDel(int x1, int x2, int y1, int y2)
	{
		//Slr p blending
		gl.glEnable(GL_BLEND);
		//Velger hvordan skal blendes
		gl.glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		//Slr p tekstur
		gl.glEnable(GL_TEXTURE_2D);
		//Binder tekstur
		gl.glBindTexture(GL_TEXTURE_2D, watertext);
		//Velger hvordan tekstur skal blandes med bakgrunn
		gl.glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
		
		//Tegner vertex'ene til vannet
		for (int x = x1; x <= x2; x++)
		{
			gl.glBegin(GL_TRIANGLE_STRIP);
			for (int y = y1; y <= y2; y++)
			{	
				gl.glNormal3f(normalsWater[x][y][0], normalsWater[x][y][1], normalsWater[x][y][2]);
				gl.glTexCoord2f(texturewater[x][y][0], texturewater[x][y][1]); gl.glVertex3f(waterhf.getLength(x,y), waterhf.getWidth(x, y), waterhf.getHeight(x, y));
				gl.glTexCoord2f(texturewater[x+1][y][0], texturewater[x+1][y][1]); gl.glVertex3f(waterhf.getLength(x+1,y), waterhf.getWidth(x+1, y), waterhf.getHeight(x+1, y));	
			}
			gl.glEnd();
		}
	
		//Slr av tekstur
		gl.glDisable(GL_TEXTURE_2D);
		//Slr av blending
		gl.glDisable(GL_BLEND);
	
	}

	//Metode som definerer material p bakken
	private void setMaterialGround()
	{
		//Hvit mye shininess
		float amb[] = {1.0f, 1.0f, 1.0f, 1.0f};
		float diff[] = {1.0f, 1.0f, 1.0f, 1.0f};
		float spec[] = {1.0f, 1.0f, 1.0f, 1.0f};
		float shine = 200f;
		
		gl.glMaterialfv(GL_FRONT, GL_AMBIENT, amb);
		gl.glMaterialfv(GL_FRONT, GL_DIFFUSE, diff);
		gl.glMaterialfv(GL_FRONT, GL_SPECULAR, spec);
		gl.glMaterialf(GL_FRONT, GL_SHININESS, shine);
	}
	
	//Metode som definerere material p vann
	private void setMaterialWater()
	{
		//Bl-grnn transparent
		float amb[] = {0.0f, 1.0f, 1.0f, 0.7f};
		float diff[] = {0.0f, 1.0f, 1.0f, 0.7f};
		float spec[] = {0.0f, 0.0f, 1.0f, 0.7f};
		float shine = 90f;
		
		gl.glMaterialfv(GL_FRONT, GL_AMBIENT, amb);
		gl.glMaterialfv(GL_FRONT, GL_DIFFUSE, diff);
		gl.glMaterialfv(GL_FRONT, GL_SPECULAR, spec);
		gl.glMaterialf(GL_FRONT, GL_SHININESS, shine);
	}

	
	//Metode som definerer material p himmel
	private void setMaterialSky()
	{
		//Hvit
		float amb[] = {1.0f, 1.0f, 1.0f, 1.0f};
		float diff[] = {1.0f, 1.0f, 1.0f, 1.0f};
		float spec[] = {1.0f, 1.0f, 1.0f, 1.0f};
		float shine = 50f;
		
		gl.glMaterialfv(GL_FRONT, GL_AMBIENT, amb);
		gl.glMaterialfv(GL_FRONT, GL_DIFFUSE, diff);
		gl.glMaterialfv(GL_FRONT, GL_SPECULAR, spec);
		gl.glMaterialf(GL_FRONT, GL_SHININESS, shine);
	}
	
	//Metode som laster egenlagd
	public int groundtext()
	{ 
		//Genererer referanse til tekstur
		gl.glGenTextures(1,texName); 
		//Binner tekstur
		gl.glBindTexture(GL_TEXTURE_2D,texName[0]); 

		//Velger hvordan tekstur skal bli gjentatt 
    gl.glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT); 
    gl.glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT); 
		
		//Velger hvordan et pixel skal rendres dersom et teksturelement 
		//dekker mer eller mindre en en pixel 
    gl.glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST); 
    gl.glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST); 

		//Henter ut verider fra egenlagd tekstur
		gl.glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, SIZE, 
		SIZE, 0, GL_RGBA, GL_FLOAT,madeTex);
		
		return texName[0];
	}
	
	//Metode som laster texture fra fil
	public int text(String fil)
	{
		//Leser teksturen fra fil og
		//inn i texture loader objektet
		texLoader.readTexture(fil);
		//Genererer referanse til tekstur
		gl.glGenTextures(1,texName); 
		
		//Binner tekstur
		gl.glBindTexture(GL_TEXTURE_2D,texName[0]); 
		
			//Velger hvordan tekstur skal bli gjentatt 
    	gl.glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_LINEAR); 
    	gl.glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_LINEAR); 
	
			//Velger hvordan et pixel skal rendres dersom et teksturelement 
			//dekker mer eller mindre en en pixel 
    	gl.glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST); 
    	gl.glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST); 
		
		//Skalerer tekstur korrekt
		texLoader.texImage2DNonScaled(true); 
		
		return texName[0];
	}
	
	//Metode som sjekker om et punkt er innenfor frustum(Kameravinkel)
	private boolean cubeInFrustum( float x, float y, float z, float size )
	{
   int p;

   for( p = 0; p < 6; p++ )
   {
      if( frustum[p][0] * (x - size) + frustum[p][1] * (y - size) + frustum[p][2] * (z - size) + frustum[p][3] > 0 )
         continue;
      if( frustum[p][0] * (x + size) + frustum[p][1] * (y - size) + frustum[p][2] * (z - size) + frustum[p][3] > 0 )
         continue;
      if( frustum[p][0] * (x - size) + frustum[p][1] * (y + size) + frustum[p][2] * (z - size) + frustum[p][3] > 0 )
         continue;
      if( frustum[p][0] * (x + size) + frustum[p][1] * (y + size) + frustum[p][2] * (z - size) + frustum[p][3] > 0 )
         continue;
      if( frustum[p][0] * (x - size) + frustum[p][1] * (y - size) + frustum[p][2] * (z + size) + frustum[p][3] > 0 )
         continue;
      if( frustum[p][0] * (x + size) + frustum[p][1] * (y - size) + frustum[p][2] * (z + size) + frustum[p][3] > 0 )
         continue;
      if( frustum[p][0] * (x - size) + frustum[p][1] * (y + size) + frustum[p][2] * (z + size) + frustum[p][3] > 0 )
         continue;
      if( frustum[p][0] * (x + size) + frustum[p][1] * (y + size) + frustum[p][2] * (z + size) + frustum[p][3] > 0 )
         continue;
      return false;
   }
   return true;
}

	//Metode som henter ut frustum
	private void extractFrustum()
	{
  
	float   proj[] = new float[16];
   float   modl[] = new float[16];
   float   clip[] = new float[16];
   float   t;

   /* Get the current PROJECTION matrix from OpenGL */
   gl.glGetFloatv( GL_PROJECTION_MATRIX, proj );

   /* Get the current MODELVIEW matrix from OpenGL */
   gl.glGetFloatv( GL_MODELVIEW_MATRIX, modl );

   /* Combine the two matrices (multiply projection by modelview) */
   clip[ 0] = modl[ 0] * proj[ 0] + modl[ 1] * proj[ 4] + modl[ 2] * proj[ 8] + modl[ 3] * proj[12];
   clip[ 1] = modl[ 0] * proj[ 1] + modl[ 1] * proj[ 5] + modl[ 2] * proj[ 9] + modl[ 3] * proj[13];
   clip[ 2] = modl[ 0] * proj[ 2] + modl[ 1] * proj[ 6] + modl[ 2] * proj[10] + modl[ 3] * proj[14];
   clip[ 3] = modl[ 0] * proj[ 3] + modl[ 1] * proj[ 7] + modl[ 2] * proj[11] + modl[ 3] * proj[15];

   clip[ 4] = modl[ 4] * proj[ 0] + modl[ 5] * proj[ 4] + modl[ 6] * proj[ 8] + modl[ 7] * proj[12];
   clip[ 5] = modl[ 4] * proj[ 1] + modl[ 5] * proj[ 5] + modl[ 6] * proj[ 9] + modl[ 7] * proj[13];
   clip[ 6] = modl[ 4] * proj[ 2] + modl[ 5] * proj[ 6] + modl[ 6] * proj[10] + modl[ 7] * proj[14];
   clip[ 7] = modl[ 4] * proj[ 3] + modl[ 5] * proj[ 7] + modl[ 6] * proj[11] + modl[ 7] * proj[15];

   clip[ 8] = modl[ 8] * proj[ 0] + modl[ 9] * proj[ 4] + modl[10] * proj[ 8] + modl[11] * proj[12];
   clip[ 9] = modl[ 8] * proj[ 1] + modl[ 9] * proj[ 5] + modl[10] * proj[ 9] + modl[11] * proj[13];
   clip[10] = modl[ 8] * proj[ 2] + modl[ 9] * proj[ 6] + modl[10] * proj[10] + modl[11] * proj[14];
   clip[11] = modl[ 8] * proj[ 3] + modl[ 9] * proj[ 7] + modl[10] * proj[11] + modl[11] * proj[15];

   clip[12] = modl[12] * proj[ 0] + modl[13] * proj[ 4] + modl[14] * proj[ 8] + modl[15] * proj[12];
   clip[13] = modl[12] * proj[ 1] + modl[13] * proj[ 5] + modl[14] * proj[ 9] + modl[15] * proj[13];
   clip[14] = modl[12] * proj[ 2] + modl[13] * proj[ 6] + modl[14] * proj[10] + modl[15] * proj[14];
   clip[15] = modl[12] * proj[ 3] + modl[13] * proj[ 7] + modl[14] * proj[11] + modl[15] * proj[15];

   /* Extract the numbers for the RIGHT plane */
   frustum[0][0] = clip[ 3] - clip[ 0];
   frustum[0][1] = clip[ 7] - clip[ 4];
   frustum[0][2] = clip[11] - clip[ 8];
   frustum[0][3] = clip[15] - clip[12];

   /* Normalize the result */
  	t = (float)Math.sqrt( (double)((frustum[0][0] * frustum[0][0]) + (frustum[0][1] * frustum[0][1]) + (frustum[0][2] * frustum[0][2]) ));
		frustum[0][0] /= t;
   frustum[0][1] /= t;
   frustum[0][2] /= t;
   frustum[0][3] /= t;

   /* Extract the numbers for the LEFT plane */
   frustum[1][0] = clip[ 3] + clip[ 0];
   frustum[1][1] = clip[ 7] + clip[ 4];
   frustum[1][2] = clip[11] + clip[ 8];
   frustum[1][3] = clip[15] + clip[12];

   /* Normalize the result */
   t = (float)Math.sqrt((double) ((frustum[1][0] * frustum[1][0]) + (frustum[1][1]) * (frustum[1][1]) + (frustum[1][2] * frustum[1][2])) );
	frustum[1][0] /= t;
   frustum[1][1] /= t;
   frustum[1][2] /= t;
   frustum[1][3] /= t;

   /* Extract the BOTTOM plane */
   frustum[2][0] = clip[ 3] + clip[ 1];
   frustum[2][1] = clip[ 7] + clip[ 5];
   frustum[2][2] = clip[11] + clip[ 9];
   frustum[2][3] = clip[15] + clip[13];

   /* Normalize the result */
   t = (float)Math.sqrt((double) ((frustum[2][0] * frustum[2][0]) + (frustum[2][1] * frustum[2][1]) + (frustum[2][2] * frustum[2][2]) ));
		frustum[2][0] /= t;
   frustum[2][1] /= t;
   frustum[2][2] /= t;
   frustum[2][3] /= t;

   /* Extract the TOP plane */
   frustum[3][0] = clip[ 3] - clip[ 1];
   frustum[3][1] = clip[ 7] - clip[ 5];
   frustum[3][2] = clip[11] - clip[ 9];
   frustum[3][3] = clip[15] - clip[13];

   /* Normalize the result */
  t = (float)Math.sqrt((double) (frustum[3][0] * frustum[3][0]) + (frustum[3][1] * frustum[3][1]) + (frustum[3][2] * frustum[3][2]) );
		frustum[3][0] /= t;
   frustum[3][1] /= t;
   frustum[3][2] /= t;
   frustum[3][3] /= t;

   /* Extract the FAR plane */
   frustum[4][0] = clip[ 3] - clip[ 2];
   frustum[4][1] = clip[ 7] - clip[ 6];
   frustum[4][2] = clip[11] - clip[10];
   frustum[4][3] = clip[15] - clip[14];

   /* Normalize the result */
   t = (float)Math.sqrt((double)( (frustum[4][0] * frustum[4][0]) + (frustum[4][1] * frustum[4][1]) + (frustum[4][2] * frustum[4][2])) );
		frustum[4][0] /= t;
   frustum[4][1] /= t;
   frustum[4][2] /= t;
   frustum[4][3] /= t;

   /* Extract the NEAR plane */
   frustum[5][0] = clip[ 3] + clip[ 2];
   frustum[5][1] = clip[ 7] + clip[ 6];
   frustum[5][2] = clip[11] + clip[10];
   frustum[5][3] = clip[15] + clip[14];

   /* Normalize the result */
   t = (float)Math.sqrt((double) ((frustum[5][0] * frustum[5][0]) + (frustum[5][1] * frustum[5][1]) + (frustum[5][2] * frustum[5][2]) ));
	frustum[5][0] /= t;
   frustum[5][1] /= t;
   frustum[5][2] /= t;
   frustum[5][3] /= t;
}

	//Klasse som behandler tastetrykk fra bruker
	class keyHandler implements KeyListener
 	{
		public void keyPressed(KeyEvent k) 
		{
			
			//Bevegelse framover
			if (k.getKeyCode() == k.VK_UP)
			{
				if (xrot <= 0) return;
				xrot -= (stride);
				repaint();
			}
			//Bevegelse bakover
			if (k.getKeyCode() == k.VK_DOWN)
			{
				if (xrot >= (hf.getSize()*stride)-stride) return;
				xrot += (stride);
				repaint();
			}
			
			//Bevegelse til venstre
			if (k.getKeyCode() == k.VK_LEFT)
			{
				if (yrot <= 0) return;
				yrot -= (stride);		
				repaint();
			}
			
			//Bevegelse til hyre
			if (k.getKeyCode() == k.VK_RIGHT)
			{
				if (yrot >= (hf.getSize()*stride)-stride) return;
				yrot += (stride);	
				repaint();
			}
			
			//Snu til venstre
			if (k.getKeyCode() == k.VK_A)
			{
				rotz -= 2.0f;
				repaint();
			}
			
			//Snu til hyre
			if (k.getKeyCode() == k.VK_Z)
			{
				rotz += 2.0f;
				repaint();
			}
			
			//Kikke opp
			if (k.getKeyCode() == k.VK_D)
			{
				lookupdown -= 1.0f;
				repaint();
			}
			
			//Kikke ned
			if (k.getKeyCode() == k.VK_C)
			{
				lookupdown += 1.0f;
				repaint();
			}
		}

		public void keyReleased(KeyEvent k) {}
		
		public void keyTyped(KeyEvent k) {}
	
	}

}

